#-*- coding: utf-8-*-

import unicodedata
import string
from fuzzywuzzy import fuzz
from fuzzywuzzy import process
import re


NUTRIMENTS = [
    {
        "key": "calories",
        "match": ["calories", "calorias", "calorie", "caloria", "calory", "kilocalories", "ener", "energy", "energia", 'kcal'],
        "name": "Calories",
        "name_es": "Calorias"
    },
    {
        "name": "Polyunsaturated Fat",
        "name_es": "Grasas Poliinsaturadas",
        "key": "poly_fat",
        "match": ["fapucis", "polyunsaturated fat", "grasas poliinsaturadas", "grasa poliinsaturada"]
    },
    {
        "name": "Monounsaturated Fat",
        "name_es": "Grasas Monoinsaturadas",
        "key": "mono_fat",
        "match": ["famscis", "monounsaturated fat", "grasas monoinsaturadas", "grasa monoinsaturada"]
    },
    {
        "name": "Trans Fat",
        "name_es": "Grasas Trans",
        "key": "trans_fat",
        "match": ["grasas trans", "trans fat", "acidos grasos trans", "trans fatty acids"]
    },
    {
        "name": "Total Fat",
        "name_es": "Grasas Totales",
        "key": "fat",
        "match": ["fat", "grasa", "grasas", "acidos grasos", "grasa total", "total fat"],
    },
    {

        "key": "niacin",
        "match": ["nia", "niacin", "niacina"],
        "name": "Niacin",
        "name_es": "Niacina"
    },
    {
        "name": "Docosahexaenoic Acid",
        "name_es": "Ácido Docosahexaenoico",
        "key": "dha",
        "match": ["docosahexaenoic", "docosahexaenoico", "dha"]
    },
    {
        "name": "Riboflavin",
        "name_es": "Riboflavina",
        "key": "riboflavin",
        "match": ["riboflavin", "riboflavina", "ribf"]
    },
    {
        "name": "Potassium",
        "name_es": "Potasio",
        "key": "k",
        "match": ["potassium", "potasio", "k"]
    },
    {
        "name": "Vitamin E",
        "name_es": "Vitamina E",
        "key": "vite",
        "match": ["vitamin e", 'vitamina e'],
    },
    {
        "name": "Iodized Salt",
        "name_es": "Yodo",
        "key": "iodine",
        "match": ["iodized_salt", "yodo", 'sal iodada']
    },
    {
        "name": "Carnitine",
        "name_es": "Carnitina",
        "key": "carnitine",
        "match": ["carnitine", "l carnitine", "carnitina"]
    },
    {
        "name": "Thiamin",
        "name_es": "Tiamina",
        "key": "tiamin",
        "match": ["thia", "thiamin", "thiamine","tiamina", "vitb1", "vitamina b1", "vitamine b1"]
    },
    {
        "name": "Saccharose",
        "name_es": "Sacarosa",
        "key": "saccharose",
        "match": ["saccharose", "sacarosa"]
    },
    {
        "name": "Vitamin C",
        "name_es": "Vitamina C",
        "key": "vitc",
        "match": ["vitamina c", "vitamin c"],
    },
    {
        "name": "Erythritol",
        "name_es": "Eritritol",
        "key": "eritritol",
        "match": ["eritritol", "erythritol"]
    },
    {
        "name": "Carbohydrate",
        "name_es": "Carbohidratos",
        "key": "carbohydrate",
        "match": ["carbohydrate", "carbohidrato", "choavl", "total carbohydrate", "carbohidratos totales"]
    },
    {
        "name": "Lactose",
        "name_es": "Lactosa",
        "key": "lactose",
        "match": ["lactose", "lactosa"]
    },
    {
        "name": "Zinc",
        "name_es": "Zinc",
        "key": "zinc",
        "match": ["zinc"]
    },
    {
        "name": "Folic Acid",
        "name_es": "Ácido Fólico",
        "key": "folic_acid",
        "match": ["folic acid", "acido folico", "acid folico", "acido folic", "acid folic", "foldfe"]
    },
    {
        "name": "Vitamin A",
        "name_es": "Vitamina A",
        "key": "vita",
        "match": ["vitamina a", "vitamin a"]
    },
    {
        "name": "Vitamin D",
        "name_es": "Vitamina D",
        "key": "vitd",
        "match": ["vitamina d", "vitamin d"]
    },
    {
        "name": "Cholesterol",
        "name_es": "Colesterol",
        "key": "cholesterol",
        "match": ["cholesterol", "colesterol"]
    },
    {
        "name": "Sugar",
        "name_es": "Azúcares",
        "key": "sugar",
        "match": ["sugar", "sugares", "azucar", "azucares", "azucares totales", "total sugars", "sugars"]
    },
    {
        "name": "Calcium",
        "name_es": "Calcio",
        "key": "calcium",
        "match": ["calcium", "calcio"]
    },
    {
        "name": "Selenium",
        "name_es": "Selenio",
        "key": "selenium",
        "match": ["selenium", "selenio"]
    },
    {
        "name": "Pantothenic Acid",
        "name_es": "Ácido Pantoténico",
        "key": "pantothenic_acid",
        "match": ["pantothenic acid", "pantac", "acido pantotenico"]
    },
    {
        "name": "Vitamin B12",
        "name_es": "Vitamina B12",
        "key": "vitb12",
        "match": ["vitb12", "vitamina b12", "vitamin b12"]
    },
    {
        "name": "Vitamin B2",
        "name_es": "Vitamina B2",
        "key": "vitb2",
        "match": ["vitb2", "vitamina b2", "vitamin b2"]
    },
    {
        "name": "Vitamin B6",
        "name_es":"Vitamina B6",
        "key": "vitb6",
        "match": ["vitb6", "vitb6", "vitamina b6", "vitamin b6"]
    },
    {
        "name": "Vitamin K",
        "name_es":"Vitamina K",
        "key": "vitk",
        "match": ["vitamina k", "vitamin k"]
    },
    {
        "name": "Fibre",
        "name_es": "Fibra",
        "key": "fibre",
        "match": ["fibtg", "fibre", "fibra", "dietary fiber", "fibra dietetica", "roughage"]
    },
    {
        "name": "Magnesium",
        "name_es": "Magnesio",
        "key": "magnesium",
        "match": ["magnessium", "magnesio"]
    },
    {
        "name": "Net Content",
        "name_es": "Contenido Neto",
        "key": "net_content",
        "match": ["net content", "netcontent", "contenido neto"]
    },
    {
        "name": "Iodide",
        "name_es": "Yoduro",
        "key": "iodide",
        "match": ["iodide", "yoduro"]
    },
    {
        "name": "Serving Quantity Information",
        "name_es": "Tamaño de porción",
        "key": "serving_quantity",
        "match": ["serving quantity", "serving quantity information", "tamano de porcion", "serv size", "serving size",
                  "serv quantity",  "serv qty"]
    },
    {
        "name": "Polyols",
        "name_es": "Polioles",
        "key": "polyols",
        "match": ["polyols", "polioles", "polyl"]
    },
    {
        "name": "Sodium Chloryde",
        "name_es": "Cloruro de Sodio",
        "key": "sodium_chloryde",
        "match": ["sodium chloryde", "cloruro de sodio", "nacl", "sales"]
    },
    {
        "name": "Saturated Fat",
        "name_es": "Grasas Saturadas",
        "key": "sat_fat",
        "match": ["fasat", "saturated fat", "grasas saturadas", "grasa saturada"]
    },
    {
        "name": "Protein",
        "name_es": "Proteína",
        "key": "protein",
        "match": ["protein", "proteina"]
    },
    {
        "name": "Galactose",
        "name_es": "Galactosa",
        "key": "galactose",
        "match": ["galactose", "galactosa"]
    },
    {
        "name": "Bicarbonate",
        "name_es": "Bicarbonato",
        "key": "bicarbonate",
        "match": ["bicarbonate", "bicarbonato", "g hc"]
    },
    {
        "name": "Iron",
        "name_es": "Hierro",
        "key": "iron",
        "match": ["hierro", "iron"]
    },
    {
        "name": "Biotin",
        "name_es": "Biotina",
        "key": "biotin",
        "match": ["biotin", "biot", "biotina"]
    },
    {
        "name": "Oleic Acid",
        "name_es": "Ácido Oleico",
        "key": "oleic_acid",
        "match": ["oleic acid", "omega 9", "acido oleico", "acid oleico", "acido oleic", "oleic acido"]
    },
    {
        "name": "Phosphorus",
        "name_es": "Fósforo",
        "key": "phosphorus",
        "match": ["phosphorus", "fosforo"]
    },
    {
        "name": "Copper",
        "name_es": "Cobre",
        "key": "copper",
        "match": ["copper", "cobre"]
    },
    {
        "name": "Sodium",
        "name_es": "Sodio",
        "key": "sodium",
        "match": ["sodio", "sodium"]
    }
]


WEIGHT_UNITS = [
    {
        "name" : "ml",
        "name_es" : "mililitro(s)",
        "match" : ["ml","mililitros","mltrs","mls","militers", "mlt"]
    },
    {
        "name" : "l",
        "name_es" : "litro(s)",
        "match" : ["l","litros","ltrs","liters", "ltr"]
    },
    {
        "name" : "mg",
        "name_es" : "miligramo(s)",
        "match" : ["miligrams","miligramos","mg","miligramo","mgs"]
    },
    {
        "name" : "ug",
        "name_es" : "microgramo(s)",
        "match" : ["micrograms","ug","miligramo", "mcg"]
    },
    {
        "name" : "g",
        "name_es" : "gramo(s)",
        "match" : ["grams","gramos","g","gramo","gr", "grm"]
    },
    {
        "name" : "onz",
        "name_es" : "onza(s)",
        "match" : ["ounce","ounces","onza","onzas","onz"]
    },
    {
        "name" : "pnt",
        "name_es" : "pinta(s)",
        "match" : ["pint","pinta","ptd", "pti"]
    },
    {
        "name" : "kg",
        "name_es" : "kilogramo(s)",
        "match" : ["kg","kgs","kilogramos","kilogramo","kilograms","kilogram","kilos", "kilo", "kgm"]
    }
]


class Nutriments:
    def __init__(self, language='en'):
        self.langage=language
        self.raw_nutriments={}
        self.nutriments={}
        self.good_name = None
        self.nutriment = False


    @staticmethod
    def create_nutriment(key=None, name=None, daily=None, unit=None, qty=None):
        return {
            "key": key,
            "name": name,
            "unit": unit,
            "qty": qty,
            "daily": daily
        }


    def guess_nutr(self, text):
        name, self.nutriment = self.get_name(text)
        if name is not False:
            self.add_nutriment(text, name=name, is_raw=True)
        else:
            self.add_nutriment(text, name='raw', is_raw=True)


    def add_nutriment(self, *args, **kargs):
        """
        This method helps you to add nutriments easily

        Individual nutriments:
            :param name: Type of nutriment. Ex: 'brand', 'ingredient', 'category'
            :param daily: The daily of the nutriment. Ex: 'Brand's Name', 'Ingredient's Name'
            :param unit: The units of the nutriment. Ex: 'Kg', 'Kilograms', 'Lt'
            :param qty: The quantity of the nutriment. Ex 1, 100, 150
            :param order: The order of the nutriment
            :return: Update the nutriments & raw_nutriments

        Batch nutriments:
            :param Dict of nutriments {'name': 'daily'}
        """
        if args:
            if len(args) == 1:
                if isinstance(args[0], dict):
                    if kargs.get('name', False) is not False:
                        for daily, qty in args[0].items():
                            self.add_nutriment(name=kargs.get('name'), qty=qty, daily=daily)
                    else:
                        for name, value in args[0].items():
                            self.add_nutriment(name=name, values=value)
                elif isinstance(args[0], str):
                    if kargs.get('name', False) is not False:
                        self.add_nutriment(name=kargs.get('name'), values=args[0])
                    else:
                       self.guess_nutr(args[0])
                elif isinstance(args[0], list) or isinstance(args[0], tuple) or isinstance(args[0], set):
                    if kargs.get('name', False) is not False:
                        if isinstance(args[0], list) or isinstance(args[0], tuple):
                            for value in args[0]:
                                if value:
                                    self.add_nutriment(name=kargs.get('name'), values=value)
                    else:
                        for value in args[0]:
                            if value:
                                self.guess_nutr(value)
                else:
                    print("ERROR adding nutrition fact: {} type is not supported ({})".format(str(args[0]), type(args[0])))
                    return False
        elif kargs:
            name = kargs.get('name', 'raw')
            values = kargs.get('values', '')
            if self.nutriment is None or self.nutriment is False:
                name_matched, self.nutriment = self.get_name(name)
            if self.nutriment is not False:
                nutr_dict = {
                    'key': self.nutriment.get('key'),
                    'name': self.nutriment.get('name_es') if self.langage == 'es' else self.nutriment.get('name')
                }
                is_raw = False
                nutr_dict.update(self.gess_values(self.nutriment,  daily=kargs.get('daily'), unit=kargs.get('unit'), qty=kargs.get('qty'), values=values))
            else:
                nutr_dict = {
                    'key': name,
                    'name': name + ' ' + values
                }
                is_raw = True
                nutr_dict.update(self.gess_values(None,  daily=kargs.get('daily'), unit=kargs.get('unit'), qty=kargs.get('qty'), values=values))




            if kargs.get('is_raw', False) is True:
                is_raw = True




            self.update_nutr(nutr_dict, is_raw)
        else:
            print("ERROR adding nutrition fact: You should pass at least 1 paramenter")


    def update_nutr(self, parsed_nutr, is_raw=True):
        key = parsed_nutr.get('key', 'raw')
        if key not in ['calories', 'serving_quantity'] and\
                (parsed_nutr.get('unit') is None or parsed_nutr.get('qty') is None):
            is_raw = True

        if is_raw is True:
            nutr = self.raw_nutriments.get(key, False)
            if nutr is False:
                self.raw_nutriments[key] = [parsed_nutr]
            elif isinstance(nutr, list):
                self.raw_nutriments[key].append(parsed_nutr)
        else:
            nutr = self.nutriments.get(key, False)
            if nutr is False:
                self.nutriments[key] = [parsed_nutr]
            elif isinstance(nutr, list):
                self.nutriments[key].append(parsed_nutr)
        self.nutriment = None


    @staticmethod
    def get_name(text, min_score=90, trusty_score=95):
        text = ''.join(x for x in unicodedata.normalize('NFKD', text) if x in string.ascii_letters or x in [" ", '1', '2' , '6', '9']).lower()
        text.strip()
        text = re.sub(r'\s+', ' ', text)
        max_score = float('-inf')
        closer_nutr = False
        for nutr in NUTRIMENTS:
            choices = nutr.get("match")
            result = process.extractOne(text, choices, scorer=fuzz.WRatio, score_cutoff=min_score)
            if result and result[1] > max_score:
                closer_nutr = nutr.get('key')
                closer_dict = nutr
                closer_dict['closer_match'] = result[0]
                max_score = result[1]
                if max_score >= trusty_score:
                    return closer_nutr, nutr
        if max_score >= min_score:
            return closer_nutr, closer_dict
        else:
            return False, False


    def gess_values(self, nutr, daily=None, unit=None, qty=None, values=None):
        guess_dict ={
            "qty": None,
            "daily": None,
            "unit": None
        }
        if qty is not False:
            if isinstance(qty, str):
                if unit is False:
                    unit, closer_unit = self.get_unit(qty)
                qty = self.str_to_float(qty)
                qty = qty if qty is not False else None

            elif isinstance(qty, float) or isinstance(qty, float):
                qty = float(qty)

        if daily is not False:
            if isinstance(daily, str):
                daily = self.str_to_float(daily)
                daily = daily if daily is not False else None

            elif isinstance(daily, float) or isinstance(daily, float):
                daily = float(daily)

        if unit is not False and isinstance(unit, str):
            unit, closer_unit = self.get_unit(unit, is_trusty=True)
            if unit is not False:
                unit = unit

        if values and isinstance(values, str):
            values_txt = ''.join(
            x for x in unicodedata.normalize('NFKD', values) if x in string.ascii_letters or x in [" ", "%"]).lower()
            if nutr is not None:
                values_txt = values_txt.replace(nutr.get('closer_match'), '')
            if unit is None:
                unit, closer_unit = self.get_unit(values_txt)
            if qty is None:
                qty_search = re.search(r'([\d\.]+) *{unit}'.format(unit=closer_unit), values)
                if qty_search:
                    try:
                        qty = float(qty_search.group(1))
                    except:
                        qty = None
                if qty is None:
                    qty_search = re.findall(r'([\d\.]+)', values)
                    if len(qty_search) == 1:
                        if '%' not in values:
                            try:
                                qty = float(qty_search[0])
                            except:
                                qty = None
                        else:
                            try:
                                daily = float(qty_search[0])
                            except:
                                daily = None
                    elif len(qty_search) == 2:
                        if '%' in values:
                            daily_search = re.search(r'([\d\.]+) *%', values)
                            if daily_search:
                                try:
                                    daily_str = daily_search.group(1)
                                    qty = set(qty_search) - {daily_search}

                                    if len(qty) == 1:
                                        try:
                                            qty = float(qty)
                                        except:
                                            qty = None
                                    daily = float(daily_str)
                                except:
                                    daily = None
            if daily is None and '%' in values:
                daily_search = re.findall(r'([\d\.]+)', values)
                if len(daily_search) == 1:
                    try:
                        daily = float(qty_search[0])
                    except:
                        daily = None
                else:
                    daily_search = re.search(r'([\d\.]+) *%', values)
                    if daily_search:
                        try:
                            daily = float(daily_search.group(1))
                        except:
                            daily = None

        if guess_dict["qty"] is None:
            guess_dict["qty"] = qty
        if guess_dict["daily"] is None:
            guess_dict["daily"] = daily
        if guess_dict["unit"] is None:
            guess_dict["unit"] = unit

        return guess_dict


    def get_unit(self, text, is_trusty=False, min_score=95, trusty_score=100):
        if not isinstance(text, str):
            return False
        text = ''.join(
            x for x in unicodedata.normalize('NFKD', text) if x in string.ascii_letters or x in [" "]).lower()
        text.strip()
        text = re.sub(r'[\d%]', '', text)
        text = re.sub(r'\s+', ' ', text)
        if is_trusty:
            trustier_text = re.search('[\d\.]* *([\(\)a-z]+)', text)
            trustier_text = trustier_text.group(1) if trustier_text else False
        else:
            trustier_text = False
        max_score = float('-inf')
        closer_unit = False
        closer_match = False
        for unit in WEIGHT_UNITS:
            choices = unit.get("match")
            if trustier_text is not False:
                result = process.extractOne(text, choices, scorer=fuzz.UWRatio, score_cutoff=min_score)
                if result is not None:
                    aux = unit.get('name_es') if self.langage == 'es' else unit.get('name')
                    return aux, result[0]
                continue
            result = process.extractOne(text, choices, scorer=fuzz.token_set_ratio, score_cutoff=min_score)
            if result and result[1] >= max_score:
                closer_unit = unit.get('name_es') if self.langage == 'es' else unit.get('name')
                closer_match = result[0]
                max_score = result[1]
            if is_trusty is True or max_score >= trusty_score:
                return closer_unit, closer_match
        if max_score > min_score:
            return closer_unit, closer_match
        else:
            return False, False


    @staticmethod
    def str_to_float(text):
        if not isinstance(text, str):
            return False
        text = re.sub(r'\s+', ' ', text)

        text = re.sub(r'[^\d\.]', '', text)
        text.strip()
        try:
            return float(text)
        except:
            return False


    def get_nutrs(self):
        return self.nutriments, self.raw_nutriments




#-*- coding: utf-8-*-
import unicodedata
import string



def key_format(data):
    '''
    Normaliza textos a formatos sin acentos o con formato
    para keys a base de datos
    '''
    if not isinstance(data, str):
        return None
    return ''.join(x for x in unicodedata.normalize('NFKD', data) if x in string.ascii_letters or x == " ").lower().replace(" ","_").replace("_y_","_").replace("_e_","_")


def clean_list(list_, remove_duplicates=True):
    if list_:
        output = [elem.strip() for elem in list_ if elem.strip()]
        if remove_duplicates is True:
            return list(set(output))
        else:
            return output
    else:
        return []





#-*- coding: utf-8-*-
import socket
import logging
from logging.handlers import SysLogHandler
import google.cloud.logging
import os

logger = None
APP_NAME = os.getenv('APP_NAME', '__name__')
LOG_LEVEL = os.getenv('LOG_LEVEL', 'DEBUG')
LOG_HOST = os.getenv('LOG_HOST', 'logs5.papertrailapp.com')
LOG_PORT = os.getenv('LOG_PORT', 27971)
ENV = os.getenv('ENV', 'LOCAL')

# Instantiates a client
client = None
try:
    client = google.cloud.logging.Client()
    # # Connects the logger to the root logging handler with log level debug or higher
    # client.setup_logging(log_level=getattr(logging,LOG_LEVEL))
except Exception:
    logging.info('Not GCP credentials')


def create_logger(name=APP_NAME):
    ''' Create logger and add handlers and filters
    '''
    global logger
    logger = logging.getLogger(name)

    log_format = '%(service)s %(hostname)s: [%(env)s] %(message)s '
    log_formatter = logging.Formatter(log_format,datefmt='%b %d %H:%M:%S')
    log_filter = ContextFilter()
    remote_handler = SysLogHandler(address=(LOG_HOST, int(LOG_PORT)))
    remote_handler.setFormatter(log_formatter)

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(log_formatter)

    if client is not None:
        hdlr = client.get_default_handler()
        # google.cloud.logging.handlers.setup_logging(hdlr)
        logger.addHandler(hdlr)
        logger.addHandler(logging.StreamHandler())
    
    logger.setLevel(LOG_LEVEL)
    logger.addFilter(log_filter)

    
    logger.addHandler(console_handler)
    logger.addHandler(remote_handler)


def get_logger():
    ''' Get the app logger by the app name
    '''
    return logging.getLogger(APP_NAME)


class ContextFilter(logging.Filter):
    hostname = socket.gethostname()
    service = APP_NAME
    env = ENV

    def filter(self, record):
        record.hostname = ContextFilter.hostname
        record.service = ContextFilter.service
        record.env = ContextFilter.env
        return True


class AppLogger(object):
    ''' General apps logger
    '''
    logger = None
    level = None
    format = '%(asctime)s %(service)s %(hostname)s %(levelname)s: %(message)s'
    filer = None
    # handlers
    console = None
    remote = None
    # remote 
    host = None
    port = None
    
    def __init__(self, level=LOG_LEVEL, host=LOG_HOST, port=LOG_PORT):
        ''' Instantiate loggers (remote and local)
            - Console handler: log to console
            - PaperTrail handler: send logs to remote host
        '''
        self.level = getattr(logging,level)
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(self.level)
        self.formatter = logging.Formatter(self.format,datefmt='%b %d %H:%M:%S')
        self.filter = ContextFilter()

        # Console Handler
        self.console = logging.StreamHandler()
        self.console.setFormatter(self.formatter)
        self.console.addFilter(self.filter)
        self.logger.addHandler(self.console)

        # Remote handler
        if host != None and port != None:
            self.host = host
            self.port = int(port)
            self.remote = SysLogHandler(address=(self.host, self.port))
            self.remote.addFilter(self.filter)
            self.remote.setFormatter(self.formatter)
            self.logger.addHandler(self.remote)

    def info(self,msg):
        ''' Info logging
        '''
        self.logger.info(msg)

    def debug(self,msg):
        ''' Debug logging
        '''
        self.logger.debug(msg)

    def warn(self,msg):
        ''' Warning logging
        '''
        self.logger.warn(msg)

    def error(self,msg):
        ''' Error logging
        '''
        self.logger.error(msg)

    def critical(self,msg):
        ''' Critical logging
        '''
        self.logger.critical(msg)


if __name__ == '__main__':
    logger = get_logger()
    logger.info('Info message sent!!')
    logger.debug('Debug message sent....')
    logger.warning('Warning message sent....')
    logger.error('Error!!! message sent....')



# -*- coding: utf-8 -*-
import logging
import pika
import os
import json
import sys
import uuid

APP_NAME = os.getenv('APP_NAME', '__name__')
LOG_LEVEL = os.getenv('LOG_LEVEL', 'DEBUG')

LOGGER = logging.getLogger(APP_NAME)
LOGGER.setLevel(LOG_LEVEL)

pika_logger = logging.getLogger('pika')
pika_logger.setLevel("ERROR")

STREAMER_ROUTING_KEY = os.getenv('STREAMER_ROUTING_KEY', 'routing')
SMONITOR_KEY = os.getenv('SMONITOR','smonitor')

if os.getenv('ENV', 'LOCAL').lower() == 'dev':
    if not '_dev' in STREAMER_ROUTING_KEY:
        STREAMER_ROUTING_KEY += '_dev'
    if not '_dev' in SMONITOR_KEY:
        SMONITOR_KEY += '_dev'

RETAILER_KEY = os.getenv('RETAILER_KEY')
SCRAPER_TYPE = os.getenv('SCRAPER_TYPE')


class RabbitEngine(object):
    """This is a consumer/producer that can handle unexpected interactions
    with RabbitMQ such as channel and connection closures.

    If RabbitMQ closes the connection, it will reopen it. You should
    look at the output, as there are limited reasons why the connection may
    be closed, which usually are tied to permission related issues or
    socket timeouts.

    If the channel is closed, it will indicate a problem with one of the
    commands that were issued and that should surface in the output as well.

    """

    def __init__(self, config, blocking=False, purge=False):
        """Setup the example publisher object, passing in the URL we will use
        to connect to RabbitMQ.
        :param str amqp_url: The URL for connecting to RabbitMQ
        """

        self._connection = None
        self._channel = None
        self._deliveries = []
        self._acked = 0
        self._nacked = 0
        self._message_number = 0
        self._stopping = False
        self._closing = False
        self._callback = self.on_message
        self._async = not blocking
        self._purge = purge

        self.USER = config['user'] if 'user' in config.keys() else os.getenv('STREAMER_USER','guest')
        self.PWD = config['password'] if 'password' in config.keys() else os.getenv('STREAMER_PASS','guest')
        self.HOST = config['host'] if 'host' in config.keys() else os.getenv('STREAMER_HOST','localhost')
        self.VHOST = config['vhost'] if 'vhost' in config.keys() else os.getenv('STREAMER_VIRTUAL_HOST','')
        self.PORT = config['port'] if 'port' in config.keys() else os.getenv('STREAMER_PORT','5672')
        self.CONN_ATTEMPTS = str(config['connection_attempts']) if 'connection_attempts' in config.keys() else '3'
        self.HEARTBEAT = config['heartbeat_interval'] if 'heartbeat_interval' in config.keys() else '0'

        self.EXCHANGE = config['exchange'] if 'exchange' in config.keys() else os.getenv('STREAMER_EXCHANGE','data')
        self.EXCHANGE_TYPE = config['exchange_type'] if 'exchange_type' in config.keys() else os.getenv('STREAMER_EXCHANGE_TYPE','direct')
        self.ROUTING_KEY = config['routing_key'] if 'routing_key' in config.keys() else os.getenv('STREAMER_ROUTING_KEY','')
        self.QUEUE = config['queue'] if 'queue' in config.keys() else os.getenv('STREAMER_QUEUE','')
        
        self._prefetch = config.get('prefetch', None)

    def set_callback(self,callback):
        self._callback = callback

    def connect(self):
        """This method connects to RabbitMQ, returning the connection handle.
        When the connection is established, the on_connection_open method
        will be invoked by pika.
        :rtype: pika.SelectConnection
        """
        if self._async:
            self._url = 'amqp://%s:%s@%s:%s/%s?connection_attempts=%s&heartbeat_interval=%s' % \
                        (self.USER,self.PWD,self.HOST,self.PORT, self.VHOST,self.CONN_ATTEMPTS,self.HEARTBEAT)
            
            LOGGER.debug('Connecting to %s', self._url)
            self._connection = pika.SelectConnection(pika.URLParameters(self._url),
                                    self.on_connection_open,
                                    stop_ioloop_on_close=False)
        else:
            self._connection = pika.BlockingConnection(
                pika.ConnectionParameters(
                                        host=self.HOST,
                                        port=int(self.PORT),
                                        credentials=pika.credentials.PlainCredentials(self.USER, self.PWD),
                                        virtual_host=self.VHOST,
                                        heartbeat_interval=0
                                        )
                                    )
            self._channel = self._connection.channel()
            self._channel.exchange_declare(exchange=self.EXCHANGE,exchange_type=self.EXCHANGE_TYPE)
            self._channel.queue_declare(queue=self.QUEUE)
            if self._purge:
                self._channel.queue_purge(queue=self.QUEUE)

    def on_blocked_connection_closed(self):
        """ Method to apply reconnection when a Blocking connection 
            has been lost, it applies the same reconnection code.
        """
        if not self._connection:
            self.connect()
        elif self._connection.is_closed:
            self.connect()

    def on_connection_open(self, unused_connection):
        """This method is called by pika once the connection to RabbitMQ has
        been established. It passes the handle to the connection object in
        case we need it, but in this case, we'll just mark it unused.
        :type unused_connection: pika.SelectConnection
        """
        LOGGER.debug('Connection opened')
        self.add_on_connection_close_callback()
        self.open_channel()

    def add_on_connection_close_callback(self):
        """This method adds an on close callback that will be invoked by pika
        when RabbitMQ closes the connection to the publisher unexpectedly.

        """
        LOGGER.debug('Adding connection close callback')
        self._connection.add_on_close_callback(self.on_connection_closed)

    def on_connection_closed(self, connection, reply_code, reply_text):
        """This method is invoked by pika when the connection to RabbitMQ is
        closed unexpectedly. Since it is unexpected, we will reconnect to
        RabbitMQ if it disconnects.

        :param pika.connection.Connection connection: The closed connection obj
        :param int reply_code: The server provided reply_code if given
        :param str reply_text: The server provided reply_text if given

        """
        self._channel = None
        if self._closing:
            self._connection.ioloop.stop()
        else:
            LOGGER.debug('Connection closed, reopening in 5 seconds: (%s) %s',
                           reply_code, reply_text)
            self._connection.add_timeout(5, self.reconnect)

    def reconnect(self):
        """Will be invoked by the IOLoop timer if the connection is
        closed. See the on_connection_closed method.

        """
        # This is the old connection IOLoop instance, stop its ioloop
        self._connection.ioloop.stop()

        if not self._closing:

            # Create a new connection
            self.connect()

            # There is now a new connection, needs a new ioloop to run
            self._connection.ioloop.start()

    def open_channel(self):
        """Open a new channel with RabbitMQ by issuing the Channel.Open RPC
        command. When RabbitMQ responds that the channel is open, the
        on_channel_open callback will be invoked by pika.

        """
        LOGGER.debug('Creating a new channel')
        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        """This method is invoked by pika when the channel has been opened.
        The channel object is passed in so we can make use of it.

        Since the channel is now open, we'll declare the exchange to use.

        :param pika.channel.Channel channel: The channel object

        """
        LOGGER.debug('Channel opened')
        self._channel = channel
        if self._prefetch:
                LOGGER.debug('Prefetch count %s', self._prefetch)
                self._channel.basic_qos(prefetch_count=int(self._prefetch))
        self.add_on_channel_close_callback()
        self.setup_exchange(self.EXCHANGE)

    def add_on_channel_close_callback(self):
        """This method tells pika to call the on_channel_closed method if
        RabbitMQ unexpectedly closes the channel.

        """
        LOGGER.debug('Adding channel close callback')
        self._channel.add_on_close_callback(self.on_channel_closed)

    def on_channel_closed(self, channel, reply_code, reply_text):
        """Invoked by pika when RabbitMQ unexpectedly closes the channel.
        Channels are usually closed if you attempt to do something that
        violates the protocol, such as re-declare an exchange or queue with
        different parameters. In this case, we'll close the connection
        to shutdown the object.

        :param pika.channel.Channel: The closed channel
        :param int reply_code: The numeric reason the channel was closed
        :param str reply_text: The text reason the channel was closed

        """
        LOGGER.debug('Channel %i was closed: (%s) %s',
                       channel, reply_code, reply_text)
        self._connection.close()

    def setup_exchange(self, exchange_name):
        """Setup the exchange on RabbitMQ by invoking the Exchange.Declare RPC
        command. When it is complete, the on_exchange_declareok method will
        be invoked by pika.

        :param str|unicode exchange_name: The name of the exchange to declare

        """
        LOGGER.debug('Declaring exchange %s', exchange_name)
        self._channel.exchange_declare(self.on_exchange_declareok,
                                       exchange_name,
                                       self.EXCHANGE_TYPE)

    def on_exchange_declareok(self, unused_frame):
        """Invoked by pika when RabbitMQ has finished the Exchange.Declare RPC
        command.

        :param pika.Frame.Method unused_frame: Exchange.DeclareOk response frame

        """
        LOGGER.debug('Exchange declared')
        self.setup_queue(self.QUEUE)

    def setup_queue(self, queue_name):
        """Setup the queue on RabbitMQ by invoking the Queue.Declare RPC
        command. When it is complete, the on_queue_declareok method will
        be invoked by pika.

        :param str|unicode queue_name: The name of the queue to declare.

        """
        LOGGER.debug('Declaring queue %s', queue_name)
        self._channel.queue_declare(callback=self.on_queue_declareok, queue=queue_name, auto_delete=False)

    def on_queue_declareok(self, method_frame):
        """Method invoked by pika when the Queue.Declare RPC call made in
        setup_queue has completed. In this method we will bind the queue
        and exchange together with the routing key by issuing the Queue.Bind
        RPC command. When this command is complete, the on_bindok method will
        be invoked by pika.

        :param pika.frame.Method method_frame: The Queue.DeclareOk frame

        """
        LOGGER.debug('Binding %s to %s with %s',
                    self.EXCHANGE, self.QUEUE, self.ROUTING_KEY)
        self._channel.queue_bind(self.on_bindok, self.QUEUE,
                                 self.EXCHANGE, self.ROUTING_KEY)

    def on_bindok(self, unused_frame):
        """Invoked by pika when the Queue.Bind method has completed. At this
        point we will start consuming messages by calling start_consuming
        which will invoke the needed RPC commands to start the process.

        :param pika.frame.Method unused_frame: The Queue.BindOk response frame

        """
        LOGGER.debug('Queue bound')
        self.start_consuming()

    def start_consuming(self):
        """This method sets up the consumer by first calling
        add_on_cancel_callback so that the object is notified if RabbitMQ
        cancels the consumer. It then issues the Basic.Consume RPC command
        which returns the consumer tag that is used to uniquely identify the
        consumer with RabbitMQ. We keep the value to use it when we want to
        cancel consuming. The on_message method is passed in as a callback pika
        will invoke when a message is fully received.

        """
        LOGGER.debug('Issuing consumer related RPC commands')
        self.add_on_cancel_callback()
        self._consumer_tag = self._channel.basic_consume(self._callback,
                                                         self.QUEUE)

    def start_publishing(self):
        """This method will enable delivery confirmations and schedule the
        first message to be sent to RabbitMQ
        """
        LOGGER.debug('Issuing consumer related RPC commands')
        self.enable_delivery_confirmations()
        #self.schedule_next_message()

    def enable_delivery_confirmations(self):
        """Send the Confirm.Select RPC method to RabbitMQ to enable delivery
        confirmations on the channel. The only way to turn this off is to close
        the channel and create a new one.

        When the message is confirmed from RabbitMQ, the
        on_delivery_confirmation method will be invoked passing in a Basic.Ack
        or Basic.Nack method from RabbitMQ that will indicate which messages it
        is confirming or rejecting.
        """
        LOGGER.debug('Issuing Confirm.Select RPC command')
        self._channel.confirm_delivery(self.on_delivery_confirmation)

    def on_delivery_confirmation(self, method_frame):
        """Invoked by pika when RabbitMQ responds to a Basic.Publish RPC
        command, passing in either a Basic.Ack or Basic.Nack frame with
        the delivery tag of the message that was published. The delivery tag
        is an integer counter indicating the message number that was sent
        on the channel via Basic.Publish. Here we're just doing house keeping
        to keep track of stats and remove message numbers that we expect
        a delivery confirmation of from the list used to keep track of messages
        that are pending confirmation.

        :param pika.frame.Method method_frame: Basic.Ack or Basic.Nack frame

        """
        confirmation_type = method_frame.method.NAME.split('.')[1].lower()
        LOGGER.debug('Received %s for delivery tag: %i',
                    confirmation_type,
                    method_frame.method.delivery_tag)
        if confirmation_type == 'ack':
            self._acked += 1
        elif confirmation_type == 'nack':
            self._nacked += 1
        self._deliveries.remove(method_frame.method.delivery_tag)
        LOGGER.debug('Published %i messages, %i have yet to be confirmed, '
                    '%i were acked and %i were nacked',
                    self._message_number, len(self._deliveries),
                    self._acked, self._nacked)

    def send(self,message):
        self._message = message
        self.publish_message(message)

    def publish_message(self,message):
        if self._stopping:
            return
        properties = pika.BasicProperties(app_id="byprice",content_type='application/json', delivery_mode=2)
        # Add reconnection when is disconnected
        self.on_blocked_connection_closed()
        self._channel.basic_publish(exchange=self.EXCHANGE,
                                    routing_key=self.ROUTING_KEY,
                                    body=json.dumps(message, ensure_ascii=False),
                                    properties=properties)
        self._message_number += 1
        LOGGER.debug('Published message # %i', self._message_number)

    def add_on_cancel_callback(self):
        """Add a callback that will be invoked if RabbitMQ cancels the consumer
        for some reason. If RabbitMQ does cancel the consumer,
        on_consumer_cancelled will be invoked by pika.

        """
        LOGGER.debug('Adding consumer cancellation callback')
        self._channel.add_on_cancel_callback(self.on_consumer_cancelled)

    def on_consumer_cancelled(self, method_frame):
        """Invoked by pika when RabbitMQ sends a Basic.Cancel for a consumer
        receiving messages.

        :param pika.frame.Method method_frame: The Basic.Cancel frame

        """
        LOGGER.debug('Consumer was cancelled remotely, shutting down: %r',
                    method_frame)
        if self._channel:
            self._channel.close()

    def on_message(self, unused_channel, basic_deliver, properties, body):
        """Invoked by pika when a message is delivered from RabbitMQ. The
        channel is passed for your convenience. The basic_deliver object that
        is passed in carries the exchange, routing key, delivery tag and
        a redelivered flag for the message. The properties passed in is an
        instance of BasicProperties with the message properties and the body
        is the message that was sent.

        :param pika.channel.Channel unused_channel: The channel object
        :param pika.Spec.Basic.Deliver: basic_deliver method
        :param pika.Spec.BasicProperties: properties
        :param str|unicode body: The message body
        """
        LOGGER.debug('Received message # %s from %s: %s',
                    basic_deliver.delivery_tag, properties.app_id, body)
        self.acknowledge_message(basic_deliver.delivery_tag)

    def acknowledge_message(self, delivery_tag):
        """Acknowledge the message delivery from RabbitMQ by sending a
        Basic.Ack RPC method for the delivery tag.

        :param int delivery_tag: The delivery tag from the Basic.Deliver frame

        """
        LOGGER.debug('Acknowledging message %s', delivery_tag)
        self._channel.basic_ack(delivery_tag)

    def stop_consuming(self):
        """Tell RabbitMQ that you would like to stop consuming by sending the
        Basic.Cancel RPC command.

        """
        if self._channel:
            LOGGER.debug('Sending a Basic.Cancel RPC command to RabbitMQ')
            self._channel.basic_cancel(self.on_cancelok, self._consumer_tag)

    def on_cancelok(self, unused_frame):
        """This method is invoked by pika when RabbitMQ acknowledges the
        cancellation of a consumer. At this point we will close the channel.
        This will invoke the on_channel_closed method once the channel has been
        closed, which will in-turn close the connection.

        :param pika.frame.Method unused_frame: The Basic.CancelOk frame

        """
        LOGGER.debug('RabbitMQ acknowledged the cancellation of the consumer')
        self.close_channel()

    def close_channel(self):
        """Call to close the channel with RabbitMQ cleanly by issuing the
        Channel.Close RPC command.

        """
        LOGGER.debug('Closing the channel')
        self._channel.close()

    def run(self):
        """Run the example consumer by connecting to RabbitMQ and then
        starting the IOLoop to block and allow the SelectConnection to operate.
        """
        self.connect()
        self._connection.ioloop.start()

    def stop(self):
        """Cleanly shutdown the connection to RabbitMQ by stopping the consumer
        with RabbitMQ. When RabbitMQ confirms the cancellation, on_cancelok
        will be invoked by pika, which will then closing the channel and
        connection. The IOLoop is started again because this method is invoked
        when CTRL-C is pressed raising a KeyboardInterrupt exception. This
        exception stops the IOLoop which needs to be running for pika to
        communicate with RabbitMQ. All of the commands issued prior to starting
        the IOLoop will be buffered but not processed.

        """
        LOGGER.debug('Stopping')
        self._closing = True
        self.stop_consuming()
        self._connection.ioloop.start()
        LOGGER.debug('Stopped')

    def close_connection(self):
        """This method closes the connection to RabbitMQ."""
        LOGGER.debug('Closing connection')
        self._connection.close()


class Streamers(object):
    stream_info = RabbitEngine({
        'queue': STREAMER_ROUTING_KEY,
        'routing_key': STREAMER_ROUTING_KEY
    }, blocking=True)

    stream_monitor = RabbitEngine({
            'queue': SMONITOR_KEY,
            'routing_key': SMONITOR_KEY
        }, blocking=True)

    def __init__(self):
        pass


    @staticmethod
    def send_stream_info(elem, param='id', first_attempt=True):

        try:
            LOGGER.debug("Streaming [{}]".format(elem.get(param, '')))
            Streamers.stream_info.send(elem)
            return True
        except Exception as e:
            LOGGER.error("Error while streaming: {}".format(e))
        if first_attempt:
            try:
                if Streamers.stream_info:
                    LOGGER.debug("Closing possible channel in stream_info")
                    Streamers.stream_info.close_channel()
                    LOGGER.debug("Closing possible connection in stream_info")
                    Streamers.stream_info.close_connection()
                else:
                    raise Exception("stream_info does not exist anymore")
            except Exception as e:
                LOGGER.error("Error while closing stream_info: {}".format(e))

            try:
                Streamers.stream_info = RabbitEngine({
                    'queue': STREAMER_ROUTING_KEY,
                    'routing_key': STREAMER_ROUTING_KEY
                }, blocking=True)
                Streamers.send_stream_info(elem, param=param, first_attempt=False)
            except Exception as e:
                LOGGER.error("Error while reconnecting streamer_info: {}".format(e))
                return False
        return False


    @staticmethod
    def send_stream_monitor(elem, param='id', first_attempt=True):
        try:
            LOGGER.debug("Streaming [{}]".format(elem.get(param, '')))
            Streamers.stream_monitor.send(elem)
            return True
        except Exception as e:
            LOGGER.error("Error while streaming: {}".format(e))

        if first_attempt:
            try:
                if Streamers.stream_monitor:
                    LOGGER.debug("Closing possible channel in stream_monitor")
                    Streamers.stream_monitor.close_channel()
                    LOGGER.debug("Closing possible connection in stream_monitor")
                    Streamers.stream_monitor.close_connection()
                else:
                    raise Exception("stream_monitor does not exist anymore")
            except Exception as e:
                LOGGER.error("Error while closing stream_monitor: {}".format(e))

            try:
                Streamers.stream_monitor = RabbitEngine({
                    'queue': SMONITOR_KEY,
                    'routing_key': SMONITOR_KEY
                }, blocking=True)
                Streamers.send_stream_monitor(elem, param=param, first_attempt=False)
            except Exception as e:
                LOGGER.error("Error while reconnecting stream_monitor: {}".format(e))
                return False
        return False


    @staticmethod
    def close_stream_info():
        Streamers.stream_info.close_channel()
        Streamers.stream_info.close_connection()


    @staticmethod
    def close_stream_monitor():
        Streamers.stream_monitor.close_channel()
        Streamers.stream_monitor.close_connection()


def stream_info(elem, param='id'):
    return Streamers.send_stream_info(elem, param)


def close_stream_info():
    Streamers.close_stream_info()


def close_stream_monitor():
    Streamers.close_stream_monitor()


def stream_monitor(signal_type, **params):
    try:
        if signal_type.lower()=='master':
            if SCRAPER_TYPE == 'item' or SCRAPER_TYPE == 'price':
                required_params = ['params', 'num_stores']
            elif SCRAPER_TYPE == 'store':
                required_params = []
            else:
                raise Exception("SCRAPER_TYPE {} is not defined for master".format(str(SCRAPER_TYPE)))
            for param in required_params:
                if params.get(param, False) is False:
                    raise Exception('{} not defined in master signal'.format(param))
            ms_id = str(uuid.uuid1())
            elem = {
                'signal': 'master',
                'ms_id': ms_id,
                'retailer_key': RETAILER_KEY,
                'params': params.get('params'),
                'num_stores': params.get('num_stores'),
                'type_': SCRAPER_TYPE
            }
            if __sm(elem):
                return ms_id
            else:
                return False

        elif signal_type.lower()=='worker':
            required_params = ['ms_id', 'step', 'store_id']
            for param in required_params:
                if params.get(param, False) is False:
                    raise Exception('{} not defined in worker signal'.format(param))
            ws_id = str(uuid.uuid1())
            elem = {
                'signal': 'worker',
                'retailer_key': RETAILER_KEY,
                'type_': SCRAPER_TYPE,
                'ms_id': params.get('ms_id'),
                'step': params.get('step'),
                'store_id': params.get('store_id'),
                'value': params.get('value'),
                'ws_id': ws_id,
                'br_stats': params.get('br_stats', {})
            }
            if __sm(elem):
                return ws_id
            else:
                return False
        elif signal_type.lower()=='error':
            if params.get('ws_id'):
                required_params = ['ws_id', 'store_id', 'code', 'reason']
                for param in required_params:
                    if params.get(param, False) is False:
                        raise Exception('{} param not defined in error signal'.format(param))
            elif params.get('ms_id'):
                required_params = ['code', 'reason']
                for param in required_params:
                    if not params.get(param):
                        raise Exception('{} not defined in error signal'.format(param))
            es_id = str(uuid.uuid1())
            elem = {
                'signal': 'error',
                'retailer_key': RETAILER_KEY,
                'type_': SCRAPER_TYPE,
                'ws_id': params.get('ws_id'),
                'ms_id': params.get('ms_id'),
                'store_id': params.get('store_id'),
                'code': params.get('code'),
                'reason': params.get('reason'),
                'es_id': es_id
            }
            if __sm(elem):
                return es_id
            else:
                return False

        else:
            raise Exception('Wrong signal_type')
    except Exception as e:
        LOGGER.error("Error parsing smonitor signal: {}".format(e))
        return False


def __sm(elem, param='signal'):
    """
    Stream the elem to smonitor queue RabbitMQ
    :param elem: dict
    :param param: str
    :return: Boolean
    """
    return Streamers.send_stream_monitor(elem, param)


class MonitorException(Exception):
    """
    Error class with attributes code and reason
    """
    def __init__(self, code=2, reason=''):
        self.code = code
        self.reason = reason




#-*- coding: utf-8-*-

import unicodedata
import string
from fuzzywuzzy import fuzz
from fuzzywuzzy import process
import re

ATTRIBUTES = [
    {
        "key": "presentation",
        "match": ["presentation", "presentacion"],
    },
    {
        "key": "measure_unit",
        "match": ["units", "measure", "measurement", "medida", "unidades"],
    },
    {
        "key": "size",
        "match": ["size", "talla", "tamano", 'siz'],
    },
    {
        "key": "material",
        "match": ["material", "mtrl"],
    },
    {
        "key": "ingredient",
        "match": ["ingredient", "ingrediente", "ingredients", "ingredientes"]
    },
    {
        "key": "category",
        "match": ["category", "categories", "categoria", "categorias"]
    },
    {
        "key": "laboratory",
        "match": ["laboratorio", "laboratory", "lab"]
    },
    {
        "key": "model",
        "match": ["model", "modelo"]
    },
    {
        "key": "brand",
        "match": ["brand", "marca"]
    },
    {
        "key": "provider",
        "match": ["proveedor", "provider"]
    },
    {
        "key": "flavour",
        "match": ["flavour", "taste", "savor", "sabor"]
    },
    {
        "key": "color",
        "match": ["color"]
    },
    {
        "key": "genre",
        "match": ["genre", "gender", "kind", "genero", "sex", "sexo"]
    },
    {
        "key": "age",
        "match": ["age", "ages", "edad", "edades"]
    },
    {
        "key": "condition",
        "match": ["condition", "condicion", "status", "estado"]
    },
    {
        "key": "height",
        "match": ["height", "alto", "hgt"]
    },
    {
        "key": "width",
        "match": ["width", "ancho", "wdth"]
    },
    {
        "key": "length",
        "match": ["length", "long", "largo"]
    },
    {
        "key": "weight",
        "match": ["weight", "peso"]
    }
]

ATTRIBUTES_FIXED = {
    "presentation": {
        "value": [
            {
                "match": ["taza", "copa", "cup"],
                "name": "cup",
                "name_es": "taza"
            },
            {
                "match": ["packs", "unidades", "cnt", "count", "units", "ct"],
                "name": "units",
                "name_es": "unidades"
            },
            {
                "match": ["dose","dosis"],
                "name": 'dose',
                "name_es": "dosis"
            },
            {
                "match": ["canister", "frasco"],
                "name": "canister",
                "name_es": "frasco"
            },
            {
                "match": ["lata", "can"],
                "name": "can",
                "name_es": "lata"
            },
            {
                "match" : ["caja", "box", "boxes"],
                "name": "box",
                "name_es": "caja"
            },
            {
                "name" : "tablet",
                "name_es" : "tabletas",
                "match" : ["tableta","tabletas","tab","tabs","grajeas","grageas","tabl"]
            },
            {
                "name" : "capsule",
                "name_es" : "cápsulas",
                "match" : ["capsula","capsulas","comprimidos","caps","cap"]
            },
            {
                "name" : "cream",
                "name_es" : "crema",
                "match" : ["crema","cream"]
            },
            {
                "name" : "unit",
                "name_es" : "unidades",
                "match" : ["unidades","unidad","units","pieza","piezas","pz","pzs","pzas", "u" , "h87", "1n", "counts"]
            },
            {
                "name" : "kit",
                "name_es" : "paquete",
                "match" : ["set", "kt", "kit", "pack", "paquete", "combo"]
            },
            {
                "name" : "ointment",
                "name_es" : "ungüento",
                "match" : ["unguento", "ung", "ointment"]
            },
            {
                "name" : "suspension",
                "name_es" : "suspensión",
                "match" : ["susp","suspension"]
            },
            {
                "name" : "suppository",
                "name_es" : "supositorio",
                "match" : ["supositorio","suppository"]
            },
            {
                "name" : "env",
                "name_es" : "sobres",
                "match" : ["sobre","sobres","envelopes"]
            },
            {
                "name" : "gel",
                "name_es" : "gel",
                "match" : ["gel"]
            },
            {
                "name" : "solution",
                "name_es" : "solución",
                "match" :["sol","solucion","solution","solución"]
            },
            {
                "name" : "drops",
                "name_es" : "gotas",
                "match" :["gotas","gota","drops"]
            },
            {
                "name" : "tube",
                "name_es" : "tubo",
                "match" : ["tubo","tub","tube"]
            },
            {
                "name" : "ampoule",
                "name_es" : "ampolletas",
                "match" : ["ampolletas","ampoules","ampolleta"]
            },
            {
                "match": ["botella", "bottle"],
                "name": "bottle",
                "name_es": "botella"
            },
            {
                "match": ["bolsa", "bag"],
                "name": "bag",
                "name_es": "bolsa"
            }
        ],
        "fixed": [('value', str)],
        "must": [('value', str)],
        "name": "Presentation",
        "name_es": "Presentación"
    },
    "measure_unit": {
        "value": [
            {
                "name" : "ml",
                "name_es" : "mililitro(s)",
                "match" : ["ml","mililitros","mltrs","mls","militers", "mlt"]
            },
            {
                "name" : "l",
                "name_es" : "litro(s)",
                "match" : ["l","litros","ltrs","liters", "ltr"]
            },
            {
                "name" : "mg",
                "name_es" : "miligramo(s)",
                "match" : ["miligrams","miligramos","mg","miligramo","mgs"]
            },
            {
                "name" : "ug",
                "name_es" : "microgramo(s)",
                "match" : ["micrograms","ug","miligramo","mcg"]
            },
            {
                "name" : "g",
                "name_es" : "gramo(s)",
                "match" : ["grams","gramos","g","gramo","gr", "grm"]
            },
            {
                "name" : "onz",
                "name_es" : "onza(s)",
                "match" : ["ounce","ounces","onza","onzas","onz"]
            },
            {
                "name" : "pnt",
                "name_es" : "pinta(s)",
                "match" : ["pint","pinta","ptd", "pti"]
            },
            {
                "name" : "bar",
                "name_es" : "bar(es)",
                "match" : ["bar","bares","bars"]
            },
            {
                "name" : "kg",
                "name_es" : "kilogramo(s)",
                "match" : ["kg","kgs","kilogramos","kilogramo","kilograms","kilogram","kilos", "kilo", "kgm"]
            },
            {
                "name" : "mm",
                "name_es" : "milimetro(s)",
                "match" : ["mm","mms","milimetros","milimetro","millimeters","millimeter", "mmt"]
            },
            {
                "name" : "cm",
                "name_es" : "centimetro(s)",
                "match" : ["cm","cms","centimetros","centimetro","centimeters","centimeter"]
            },
            {
                "name": "in",
                "name_es": "pulgada(s)",
                "match": ["inch", '"', "pulgada", "in", "pulgadas"]
            },
            {
                "name" : "m",
                "name_es" : "metro(s)",
                "match" : ["m","ms","mtr"]
            },
            {
                "name" : "km",
                "name_es" : "kilometro(s)",
                "match" : ["km","kms","kilometros","kilometro","kilometers","kilometer", "kmt"]
            },
            {
                "name" : "par",
                "name_es" : "par(es)",
                "match" : ["par", "pares", "pair", "pairs", "pr"]
            }
        ],
        "fixed": [('value', str)],
        "must": [('value', str)],
        "name": "Measure Units",
        "name_es": "Unidades de Medida"
    },
    "size": {
        "value": [
            {
                "name" : "sm",
                "name_es" : "chico",
                "match" : ["ch","chico","pequeno","pequeño","small","sm","chicos","petite", 's']
            },
            {
                "name" : "md",
                "name_es" : "mediano",
                "match" : ["md","mediano","medium","medianos", 'm']
            },
            {
                "name" : "lg",
                "name_es" : "grande",
                "match" : ["lg","large","grande","grandes", 'l']
            },
            {
                "name" : "xl",
                "name_es" : "extra grande",
                "match" : ["extragrande","extra grande","extra large","xl","xxl","xxxl","extralarge"]
            }
        ],
        "fixed": [('value', str)],
        "must": [('value', str)],
        "type": str,
        "name": "Size",
        "name_es": "Tamaño"
    },
    "material": {
       "name": "Material",
       "name_es": "Material",
       "must": [('value', str)],
    },
    "ingredient": {
        "name": "Ingredient",
        "name_es": "Ingrediente",
        "must": [('value', str)]
    },
    "category":   {
        "name": "Category",
        "name_es": "Categoría",
        "must": [('value', str), ("order", int)],
    },
    "laboratory": {
        "name": "Laboratory",
        "name_es": "Laboratorio",
        "must": [('value', str)]
    },
    "model": {
       "name": "Model",
       "name_es": "Modelo",
       "must": [('value', str)]
    },
    "brand": {
        "name": "Brand",
        "name_es": "Marca",
        "must": [('value', str)]
    },
    "provider": {
        "name": "Provider",
        "name_es": "Proveedor",
        "must": [('value', str)]
    },
    "flavour": {
        "name": "Flavour",
        "name_es": "Sabor",
        "must": [('value', str)]
    },
    "color": {
        "name": "Color",
        "name_es": "Color",
        "must": [('value', str)]
    },
    "genre": {
        "name": "Genre",
        "name_es": "Género",
        "must": [('value', str)]
    },
    "age": {
        "name": "Age",
        "name_es": "Edad",
        "must": [('value', str)]
    },
    "condition": {
        "name": "Condition",
        "name_es": "Condición",
        "must": [('value', str)]
    },
    "height": {
        "name": "Height",
        "name_es": "Alto",
        "unit": [
            {
                "name": "mm",
                "name_es": "milimetro(s)",
                "match": ["mm", "mms", "milimetros", "milimetro", "millimeters", "millimeter", "mmt"]
            },
            {
                "name": "cm",
                "name_es": "centimetro(s)",
                "match": ["cm", "cms", "centimetros", "centimetro", "centimeters", "centimeter"]
            },
            {
                "name": "in",
                "name_es": "pulgada(s)",
                "match": ["inch", '"', "pulgada", "in", "pulgadas"]
            },
            {
                "name": "m",
                "name_es": "metro(s)",
                "match": ["m", "ms", "mtr"]
            },
            {
                "name": "km",
                "name_es": "kilometro(s)",
                "match": ["km", "kms", "kilometros", "kilometro", "kilometers", "kilometer", "kmt"]
            }
        ],
        "fixed": [("unit", str)],
        "must": [('value', float), ("unit", str)]
    },
    "width": {
       "name": "Width",
       "name_es": "Ancho",
       "unit": [
            {
                "name": "mm",
                "name_es": "milimetro(s)",
                "match": ["mm", "mms", "milimetros", "milimetro", "millimeters", "millimeter", "mmt"]
            },
            {
                "name": "cm",
                "name_es": "centimetro(s)",
                "match": ["cm", "cms", "centimetros", "centimetro", "centimeters", "centimeter"]
            },
            {
                "name": "in",
                "name_es": "pulgada(s)",
                "match": ["inch", '"', "pulgada", "in", "pulgadas"]
            },
            {
                "name": "m",
                "name_es": "metro(s)",
                "match": ["m", "ms", "mtr"]
            },
            {
                "name": "km",
                "name_es": "kilometro(s)",
                "match": ["km", "kms", "kilometros", "kilometro", "kilometers", "kilometer", "kmt"]
            }
        ],
       "fixed": [("unit", str)],
       "must": [('value', float), ("unit", str)]
    },
    "length": {
        "name": "Length",
        "name_es": "Largo",
        "unit": [
            {
                "name": "mm",
                "name_es": "milimetro(s)",
                "match": ["mm", "mms", "milimetros", "milimetro", "millimeters", "millimeter", "mmt"]
            },
            {
                "name": "cm",
                "name_es": "centimetro(s)",
                "match": ["cm", "cms", "centimetros", "centimetro", "centimeters", "centimeter"]
            },
            {
                "name": "in",
                "name_es": "pulgada(s)",
                "match": ["inch", '"', "pulgada", "in", "pulgadas"]
            },
            {
                "name": "m",
                "name_es": "metro(s)",
                "match": ["m", "ms", "mtr"]
            },
            {
                "name": "km",
                "name_es": "kilometro(s)",
                "match": ["km", "kms", "kilometros", "kilometro", "kilometers", "kilometer", "kmt"]
            }
        ],
        "fixed": ["unit"],
        "must": ["value", "unit"],
        "type": float
    },
    "weight": {
        "name": "Weight",
        "name_es": "Peso",
        "unit": [
            {
                "name" : "ml",
                "name_es" : "mililitro(s)",
                "match" : ["ml","mililitros","mltrs","mls","militers", "mlt"]
            },
            {
                "name" : "l",
                "name_es" : "litro(s)",
                "match" : ["l","litros","ltrs","liters", "ltr"]
            },
            {
                "name" : "mg",
                "name_es" : "miligramo(s)",
                "match" : ["miligrams","miligramos","mg","miligramo","mgs" ]
            },
            {
                "name" : "ug",
                "name_es" : "microgramo(s)",
                "match" : ["micrograms","ug","miligramo", "mcg"]
            },
            {
                "name" : "g",
                "name_es" : "gramo(s)",
                "match" : ["grams","gramos","g","gramo","gr", "grm"]
            },
            {
                "name" : "onz",
                "name_es" : "onza(s)",
                "match" : ["ounce","ounces","onza","onzas","onz"]
            },
            {
                "name" : "pnt",
                "name_es" : "pinta(s)",
                "match" : ["pint","pinta","ptd", "pti"]
            },
            {
                "name" : "kg",
                "name_es" : "kilogramo(s)",
                "match" : ["kg","kgs","kilogramos","kilogramo","kilograms","kilogram","kilos", "kilo", "kgm"]
            }
        ],
        "fixed": [("unit", str)],
        "must": [('value', float), ("unit", str)]
    }
}

class Attributes:
    def __init__(self, language='en'):
        self.langage=language
        self.raw_attributes={}
        self.attributes={}

    @staticmethod
    def create_attribute(key=None, name=None, value=None, unit=None, qty=None, order=None):
        return {
            "key": key,
            "name": name,
            "value": value,
            "unit": unit,
            "qty": qty,
            "order": order
        }

    def guess_attr(self, text):
        name = self.get_name(text)
        if name is not False:
            self.add_attribute(text, name=name, is_raw=True)
        else:
            self.add_attribute(text, name='raw', is_raw=True)

    def add_attribute(self, *args, **kargs):
        """
        This method helps you to add attributes easily

        Individual attributes:
            :param name: Type of attribute. Ex: 'brand', 'ingredient', 'category'
            :param value: The value of the attribute. Ex: 'Brand's Name', 'Ingredient's Name'
            :param unit: The units of the attribute. Ex: 'Kg', 'Kilograms', 'Lt'
            :param qty: The quantity of the attribute. Ex 1, 100, 150
            :param order: The order of the attribute
            :return: Update the attributes & raw_attributes

        Batch attributes:
            :param Dict of attributes {'name': 'value'}
        """
        if args:
            if len(args) == 1:
                if isinstance(args[0], dict):
                    if kargs.get('name', False) is not False:
                        for value, qty in args[0].items():
                            self.add_attribute(name=kargs.get('name'), value=value, qty=qty)
                    else:
                        for name, value in args[0].items():
                            self.add_attribute(name=name, value=value)
                elif isinstance(args[0], str):
                    if kargs.get('name', False) is not False:
                        self.add_attribute(value=args[0], qty=args[0], unit=args[0], name=kargs.get('name'), is_raw=kargs.get('is_raw', False))
                    else:
                       self.guess_attr(args[0])
                elif isinstance(args[0], list) or isinstance(args[0], tuple) or isinstance(args[0], set):
                    if kargs.get('name', False) is not False:
                        if isinstance(args[0], list) or isinstance(args[0], tuple):
                            for index, value in enumerate(args[0]):
                                if value:
                                    self.add_attribute(value=value,  name=kargs.get('name'), order=index+1)
                        else:
                            for value in args[0]:
                                if value:
                                    self.add_attribute(value=value, qty=value, unit=value, name=kargs.get('name'))
                    else:
                        for value in args[0]:
                            if value:
                                self.guess_attr(value)
                else:
                    print("ERROR adding attribute: {} type is not supported ({})".format(str(args[0]), type(args[0])))
                    return False
        elif kargs:
            name = kargs.get('name', 'raw')
            good_name = self.get_name(name)
            if good_name is not False:
                name = good_name
                is_raw = False
            else:
                is_raw = True
            attributes = [('value', str), ('unit', str), ('qty', float), ('order', str)]

            attr = ATTRIBUTES_FIXED.get(name, {})
            if len(attr) == 0:
                is_raw = True

            attr_dict = {
                            'key': name,
                            'name': attr.get('name_es', name) if self.langage == 'es' else attr.get('name', name)
                        }
            fixed = {attr_fixed[0]: attr_fixed[1] for attr_fixed in attr.get('fixed', [])}
            must = {attr_must[0]: attr_must[1] for attr_must in attr.get('must', [])}

            for attribute, default_type in attributes:
                if attribute in fixed.keys() and kargs.get(attribute, False) is not False:
                    for fixed_values in attr.get(attribute):
                        match = fixed_values.get('match')
                        match_found = fixed_values.get('name_es') if self.langage == 'es' else fixed_values.get('name')
                        result = self.get_attr(kargs.get(attribute, False), match, expected_type=fixed.get(attribute))
                        if result is True:
                            result = match_found
                            break
                    else:
                        result = kargs.get(attribute)
                        is_raw = True
                elif attribute in must.keys() and kargs.get(attribute, False) is not False:
                    can_be_numeric = attr_dict['key'] in ['brand', 'provider', 'laboratory', 'model']
                    result = self.check_attr(kargs.get(attribute), expected_type=must.get(attribute), can_be_numeric=can_be_numeric)
                    if result is None:
                        result = kargs.get(attribute)
                        is_raw = True
                else:
                    result = None
                    # result = self.check_attr(kargs.get(attribute), expected_type=default_type)
                    # if result is None:
                    #     result = kargs.get(attribute)

                attr_dict[attribute] = result

            if kargs.get('is_raw', False) is True:
                is_raw = True

            self.update_attr(attr_dict, is_raw)
        else:
            print("ERROR adding attribute: You should pass at least 1 paramenter")

    def update_attr(self, parsed_attr, is_raw=True):
        if is_raw is True:
            key = parsed_attr.get('key', 'raw')
            attr = self.raw_attributes.get(key, False)
            if attr is False:
                self.raw_attributes[key] = [parsed_attr]
            elif isinstance(attr, list):
                self.raw_attributes[key].append(parsed_attr)
        else:
            key = parsed_attr.get('key', 'raw')
            attr = self.attributes.get(key, False)
            if attr is False:
                self.attributes[key] = [parsed_attr]
            elif isinstance(attr, list):
                self.attributes[key].append(parsed_attr)

    @staticmethod
    def get_name(text, min_score=80, trusty_score=90):
        text = ''.join(x for x in unicodedata.normalize('NFKD', text) if x in string.ascii_letters or x in [" ", '"', "'", '.', ',']).lower()
        text = text.strip()
        text = re.sub(r'\s+', ' ', text)
        max_score = float('-inf')
        closer_attr = False
        for attr in ATTRIBUTES:
            choices = attr.get("match")
            result = process.extractOne(text, choices, scorer=fuzz.partial_token_set_ratio, score_cutoff=min_score)
            if result and result[1] > max_score:
                closer_attr = attr.get('key')
                max_score = result[1]
            if max_score > trusty_score:
                return closer_attr
        if max_score > min_score:
            return closer_attr
        else:
            return False

    @staticmethod
    def get_attr(value, choices, min_score=90, expected_type=str):
        if isinstance(value, str):
            if expected_type is str:
                text = ''.join(x for x in unicodedata.normalize('NFKD', value) if x in string.ascii_letters or x in [" ", '"', "'", '.', ',']).lower()
                text = text.strip()
                text = re.sub(r'\s+', ' ', text)
                result = process.extractOne(text, choices, scorer=fuzz.UWRatio, score_cutoff=min_score)
                if result and result[1] > min_score:
                    return True
                else:
                    return False
            elif expected_type is int or expected_type is float:
                value = re.sub(r'[^.,\s\d]', '', value)
                value = re.sub(r'\s+', ' ', value)
                value = value.strip()
                try:
                    return expected_type(value)
                except:
                    return False

        elif type(value) is expected_type:
            return value
        else:
            return False

    @staticmethod
    def check_attr(value, expected_type=str, can_be_numeric=False):
        if isinstance(value, str):
            if expected_type is str:
                if can_be_numeric == False:
                    text = ''.join(x for x in unicodedata.normalize('NFKD', value) if
                                   x in string.ascii_letters or x in [" ", '"', "'", '.', ',']).lower()
                else:
                    text = ''.join(x for x in unicodedata.normalize('NFKD', value) if
                                    x in string.ascii_letters
                                    or x in [" ", '"', "'", '.', ',']
                                    or x in string.digits
                                ).lower()
                text.strip()
                text = re.sub(r'\s+', ' ', text)
                text = text.strip()
                if text:
                    return text
                else:
                    return None
            elif expected_type is int or expected_type is float:
                value = re.sub(r'[^.,\s\d]', '', value)
                value = re.sub(r'\s+', ' ', value)
                value = value.strip()
                try:
                    return expected_type(value)
                except:
                    return None

        elif type(value) is expected_type:
            return value
        else:
            return None

    def get_attrs(self):
        return self.attributes, self.raw_attributes


import setuptools
from setuptools import find_packages


setuptools.setup(
    name = 'ByHelpers',
    packages = ['ByHelpers'], # this must be the same as the name above
    version = '0.1.6',
    description = 'Library used by ByPrice scrapers',
    author = 'Kevin B. Garcia Alonso',
    author_email = 'kevangy@hotmail.com',
    url = 'https://github.com/ByPrice/ByHelpers', # use the URL to the github repo
    download_url = 'https://github.com/ByPrice/ByHelpers/',
    keywords = ['ByPrice'],
    install_requires=[
        'pika==0.11.0',
        'fuzzywuzzy==0.16.0',
        'google-cloud-logging==1.10.0'
    ],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ),
)
